#
# Copyright 2011 Eigenlabs Ltd.  http://www.eigenlabs.com
#
# This file is part of EigenD.
#
# EigenD is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# EigenD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EigenD.  If not, see <http://www.gnu.org/licenses/>.
#

import piw
from pi import agent, domain, bundles, atom, resource
#TODO: why is the version not taken from the concrete classes, this line should not be necessary...
from plg_midiilluminator import tau_midi_light_controller_version as version
import csv
import re
#from number import Number

#
# midiilluminator_plg_native is generated by the build system through the midiilluminator_plg.pip file.
#
from midiilluminator_plg_native import midiilluminator

#
# The midiilluminator lights agent's main task is to highlight the leds on keys according
# to incoming midi note events. Depending on midi channel the note events can be displayed
# in different colors. There exist three variants for Pico, Alpha and Tau.
# Additionally there are two "direct" illumination features:
# Direct keys which allow to specifically define a color for a key.
# And direct notes which allows to highlight keys that are currently mapped to specific notes.
# (This can be helpful to illuminate certain notes as marker lights for guidance)
# The layout is currently directly represented via atoms of the midi illuminator.
# up to 20 layouts can be created, one of them can be chosen as the active one
# that is used for midi to key mapping and direct notes mapping.
# Midi and direct notes mapping can additionally be influenced by a transpose factor (in semitones).
# If midi notes, direct notes and direct keys define different colors for the same key
# a changable priority defines which light source wins.
# All configuration information can be imported and exported as Comma Separated Value files (csv),
# which are written in Excel compatible style, import supports a number of common CSV variants
# (e.g. successfully tested with Excel and LibreOffice Calc). The Pico, Alpha and Tau
# versions have separate configuration files, so they can be used in parallel.
# The config CSV files are stored in the Eigenlabs user folder in a "midi_illuminator" subfolder
# Beside configuration via CSV file five Stage tabs exist where every parameter can also be
# controlled.
#
# The current version of the midi illuminator is supposed to be connected between the
# midi input and the keyboard agent (so currently it is mainly intended for the case of a
# kgroup that covers the entire keyboard. Later versions might connect to the kgroup instead
# and take layout and scale informations from kgroups and scalers into consideration -
# currently those are completely ignored, only the layouts in the CSV file/setup file count)
# Most essential setup steps:
# * copy midiilluminator_plg_1_4_8_testing.dll and midiilluminator_plg_native.pyd into the EigenD bin directly
# * copy the "midi_illuminator" directory with the CSV files into the Eigenlabs user directory
#   (where the user side "Scale Manger" and "Scripts" folders reside)
# * enter the following Belcanto commands (replace "alpha" with "tau" or "pico" for the other instruments):
#   when starting from a blank setup you first need these two additional lines:
#     alpha manager create
#     midi input create  
#   Then it contiues similarly for a blank and a standard setup:
#     alpha midi light controller create
#     plumber hey midi input to alpha midi light controller 1 connect
#     plumber hey alpha midi light controller 1 to alpha keyboard 1 connect
# * Then choose the midi input (Eigenbrowser has to be started and port selected there afterwards):
#     eigenbrowser hey midi input port browse
#     (If based on a blank setup it might be necessary to write "midi input 1" instead of just
#      "midi input" in all commands above)
# * Now the stage tabs can be loaded (AlphaTabs.els, TauTabs.els or PicoTabs.els)
# * now the Midi illuminator is operative
# * additional setup steps are thinkable:
# - layouts can be defined via Stage tab or CSV file
# - midiilluminator layout change and transpose can be added to the talkers that
#   change scales or do transposition, so the midi illuminator wouldn't have to be
#   reconfigured manually when changing layout/scale/transpose factor
class IlluminatorAgent(agent.Agent):
    def __init__(self, address, ordinal, num_keys, name):
        agent.Agent.__init__(self, signature=version, names=name, ordinal=ordinal)
        self.__num_keys = num_keys
        self.domain = piw.clockdomain_ctl()
        self.domain.set_source(piw.makestring('*',0))
        self.__idx = {
            'outputs' : {'id' : 2, 'light output' : 1},
            'inputs' : {'id' : 1, 'midi input' : 1},
            'transpose' : 3,
            'current layout' : 4,
            "enable inputs green" : {'id' : 5, 'direct keys' : 1, 'direct notes' : 2, 'midi' : 3},
            "enable inputs orange" : {'id' : 6, 'direct keys' : 1, 'direct notes' : 2, 'midi' : 3},
            "enable inputs red" : {'id' : 7, 'direct keys' : 1, 'direct notes' : 2, 'midi' : 3},
            'keys' : 8,
            'notes' : 9,
            'midi channels' : 10,
            'key to note maps' : {'id' : 11, 'current layout' : 1, 'layout 1' : 2, 'layout 2' : 3, 'layout 3' : 4,
                                  'layout 4' : 5, 'layout 5' : 6, 'layout 6' : 7, 'layout 7' : 8, 'layout 8' : 9, 
                                  'layout 9' : 10, 'layout 10' : 11, 'layout 11' : 12, 'layout 12' : 13, 'layout 13' : 14, 
                                  'layout 14' : 15, 'layout 15' : 16, 'layout 16' : 17, 'layout 17' : 18, 'layout 18' : 19, 
                                  'layout 19' : 20, 'layout 20' : 21
                                  },
            'enable' : 12,
            'priorities' : {'id' : 13, 'direct keys' : 1, 'direct notes' : 2, 'midi' : 3},
            'export csv' : 14,
            'import csv' : 15
        }
        self.__lite_idx = {
            'outputs' : {'id' : 2, 'light output' : 1},
            'inputs' : {'id' : 1, 'midi input' : 1},
            'transpose' : 3,
            'current layout' : 4,
            "enable inputs green" : {'id' : 5, 'direct keys' : 1, 'direct notes' : 2, 'midi' : 3},
            "enable inputs orange" : {'id' : 6, 'direct keys' : 1, 'direct notes' : 2, 'midi' : 3},
            "enable inputs red" : {'id' : 7, 'direct keys' : 1, 'direct notes' : 2, 'midi' : 3},
            'enable' : 12,
            'export csv' : 14,
            'import csv' : 15
        }
        self._lite_slots = {}
        self.set_idx('outputs', atom.Atom(names="outputs"))
        self.set_idx_sub('outputs','light output', bundles.Output(1,False,names='light output'))
        self.output = bundles.Splitter(self.domain,*self.idx('outputs').values())
        self.midiilluminator = midiilluminator(self.domain, self.output.cookie(), self.__num_keys)
        self.data_input = bundles.VectorInput(self.midiilluminator.cookie(), self.domain, signals=(1,))
        self.set_idx('inputs', atom.Atom(names="inputs"))
        self.set_idx_sub('inputs', 'midi input', atom.Atom(domain=domain.BoundedFloat(-1,1), policy=self.data_input.vector_policy(1,False), names='midi input'))
        self.set_idx('transpose', atom.Atom(domain=domain.BoundedInt(-120,120), init=0, policy=atom.default_policy(self.__set_transpose), names='transpose'))
        self.set_idx('import csv', atom.Atom(domain=domain.Bool(),
                     init=False,
                     policy=atom.default_policy(self.__import),
                     names="import"))
        self.set_idx('export csv', atom.Atom(domain=domain.Bool(),
             init=False,
             policy=atom.default_policy(self.__export),
             names="export"))
        self.set_idx('current layout', atom.Atom(domain=domain.BoundedInt(1,self._num_layouts()), init=1, policy=atom.default_policy(self.__set_layout), names='current layout'))
        self.__LIGHT_NAME = ["green", "orange", "red"] 
        for color in range(0,3): # green, orange, red
            self.set_idx('enable inputs ' + self.__LIGHT_NAME[color], atom.Atom(names=self.__LIGHT_NAME[color] + " lights"))
            self.set_idx_sub('enable inputs ' + self.__LIGHT_NAME[color],'direct keys',
                         atom.Atom(domain=domain.Bool(),
                                   init=True,
                                   # scoping in Python is lexical, so we have to create an explicit scope with a second lambda
                                   # to have the expected variable values in the closure
                                   policy=atom.default_policy((lambda color: lambda state: self.__set_enable_direct_keys(color, state))(color)),
                                   names='enable direct keys'))
            self.set_idx_sub('enable inputs ' + self.__LIGHT_NAME[color],'direct notes',
                         atom.Atom(domain=domain.Bool(),
                                   init=True,
                                   policy=atom.default_policy((lambda color: lambda state: self.__set_enable_direct_notes(color, state))(color)),
                                   names='enable direct notes'))
            self.set_idx_sub('enable inputs ' + self.__LIGHT_NAME[color],'midi',
                         atom.Atom(domain=domain.Bool(),
                                   init=True,
                                   policy=atom.default_policy((lambda color: lambda state: self.__set_enable_midi(color, state))(color)),
                                   names='enable midi'))
        self.set_idx('enable', atom.Atom(domain=domain.Bool(),
                         init=True,
                         policy=atom.default_policy(self.__set_enable),
                         names="enable"))
        if self.is_lite() == False:
            self.set_idx('keys', atom.Atom(names="keys"))
            for key in range(1,self.__num_keys + 1):
                self.idx('keys')[key] = atom.Atom(domain=domain.BoundedInt(0,3), #0: black, 1: green, 2: orange, 3: red
                                       init=0,
                                       policy=atom.default_policy((lambda key: lambda val: self.__set_key(key,val))(key)),
                                       names='key ' + `key`)
            self.set_idx('notes', atom.Atom(names="notes"))
            notenum = 0
            for octave in range(0,11):
                for note in ['c', 'cis', 'd', 'dis', 'e', 'f', 'fis', 'g', 'gis', 'a', 'ais', 'b']:
                    if notenum > 127:
                        break
                    notenum = notenum + 1
                    self.idx('notes')[notenum] = atom.Atom(domain=domain.BoundedInt(0,3), #0: black, 1: green, 2: orange, 3: red
                                               init=0,
                                               policy=atom.default_policy((lambda notenum: lambda val: self.__set_note(notenum,val))(notenum)),
                                               names=`notenum` + ': ' + note + `octave`)
            self.set_idx('midi channels', atom.Atom(names = "midi channels"))
            for channel in range(1,17):
                self.idx('midi channels')[channel] = atom.Atom(domain=domain.BoundedInt(0,3), #0: black, 1: green, 2: orange, 3: red
                                              init=0,
                                              policy=atom.default_policy((lambda channel: lambda color: self.__set_channel_color(channel, color))(channel)),
                                              #policy=atom.default_policy(self.__dummy),
                                              names='color for channel ' + `channel`)
            self.set_idx('key to note maps', atom.Atom(names="key to note maps"))
            self.set_idx_sub('key to note maps', 'current layout', atom.Atom(names="current layout"))
            for key in range(1,self.__num_keys + 1):
                self.idx('key to note maps', 'current layout')[key] = atom.Atom(domain=domain.BoundedInt(0,128), #0: no note, 1-128: note number
                                      init=self._default_note(key),
                                      policy=atom.default_policy((lambda key: lambda notenum: self.__set_note_for_key(1, key, notenum))(key)),
                                      names='key ' + `key`)
            for layout in range(2,self._num_layouts() + 2):
                 self.idx('key to note maps')[layout] = atom.Atom(names="layout " + `layout - 1`)
                 for key in range(1,self.__num_keys + 1):
                         self.idx('key to note maps')[layout][key] = atom.Atom(domain=domain.BoundedInt(0,128), #0: no note, 1-128: note number
                                   init=self._default_note(key),
                                   policy=atom.default_policy((lambda layout, key: lambda notenum: self.__set_note_for_key(layout, key, notenum))(layout, key)),
                                   names='key ' + `key`)
            self.set_idx('priorities', atom.Atom(names="priorities"))
            self.set_idx_sub('priorities','direct keys', atom.Atom(domain=domain.BoundedInt(1,4),
                                init=3,
                                policy=atom.default_policy(self.__set_direct_keys_prio),
                                names='direct keys prio'))
            self.set_idx_sub('priorities','direct notes',  atom.Atom(domain=domain.BoundedInt(1,4),
                                init=2,
                                policy=atom.default_policy(self.__set_direct_notes_prio),
                                names='direct notes prio'))
            self.set_idx_sub('priorities','midi', atom.Atom(domain=domain.BoundedInt(1,4),
                                init=1,
                                policy=atom.default_policy(self.__set_midi_prio),
                                names='midi prio'))
            self.update_native()
        
    #STAGELAG WORKAROUND BEGIN
    #def _get_or_create_key_for_layout(self, layout, key):
    #    if not layout in self._layouts:
    #        self._layouts[layout] = {}
    #        self._layouts[layout][key] = self._default_note(key)
    #    elif not key in self._layouts[layout]:
    #        self._layouts[layout][key] = self._default_note(key)
    #    return self._layouts[layout][key]
    #STAGELAG WORKAROUND END
        
    class lite_slot:
        def __init__(self, value = 0):
            self._value = value
        def get_value(self):
            return self._value
        def set_value(self, value):
            self._value = value
        
    def idx(self, index, subindex = -1):
        assert index in self.__idx
        idx_obj = None  
        index_num = self.__idx[index]
        id = None
        if isinstance(index_num, dict):
            id = index_num['id']
        else:
            #assert isinstance(index_num, Number)
            id = index_num
        if subindex == -1:
            if self.is_lite() == True and not index in self.__lite_idx:
                idx_obj = self._lite_slots[id]
            else:
                idx_obj = self[id]
        else:
            assert subindex in self.__idx[index]
            if self.is_lite() == True and not index in self.__lite_idx:
                idx_obj = self._lite_slots[id][subindex]
            else:
                idx_obj = self[self.__idx[index]['id']][self.__idx[index][subindex]]
        return idx_obj
        
    def set_idx(self, index, value):
        assert index in self.__idx
        index_num = self.__idx[index]
        id = None
        if isinstance(index_num, dict):
            id = index_num['id']
        else:
            #assert isinstance(index_num, Number)
            id = index_num
        if self.is_lite() == True and not index in self.__lite_idx:
            self._lite_slots[id] = value
        else:
            self[id] = value
            
    def set_idx_sub(self, index, subindex, value):
        assert index in self.__idx
        assert subindex in self.__idx[index]
        if self.is_lite() == True and not index in self.__lite_idx:
            self._lite_slots[self.__idx[index]['id']][self.__idx[index][subindex]] = value
        else:
            self[self.__idx[index]['id']][self.__idx[index][subindex]] = value
        
    def agent_postload(self):
        self.__import(True)
        if self.idx('enable').get_value() == True:
            self.__set_enable(False)
            self.__set_enable(True)
        return True
                             
    def _default_note(self, key):
        pass

    def _num_layouts(self):
        pass
    
    def is_lite(self):
        return True
  
    def __set_transpose(self,t):
        self.idx('transpose').set_value(t)
        self.midiilluminator.set_transpose(t)
        return True
        
    def __set_layout(self, new_layout, force=False):
        old_layout = self.idx('current layout').get_value()
        if new_layout <> old_layout or force:
            self.idx('current layout').set_value(new_layout)
            dest_slot = self.idx('key to note maps','current layout')
            src_slot = self.idx('key to note maps')
            for key in range(1,self.__num_keys + 1):
                dest_slot[key].set_value(src_slot[new_layout + 1][key].get_value())
            repaint = False
            for key in range(1,self.__num_keys + 1):
                if key == self.__num_keys:
                    repaint = True
                #print "FSTDEBUG dest_slot[" + `key` + "] = " + `dest_slot[key].get_value()`
                self.midiilluminator.set_note_for_key(key, dest_slot[key].get_value(),repaint)
        return True
    
    def __set_enable_direct_keys(self, color, enable):
        slot = self.idx('enable inputs ' + self.__LIGHT_NAME[color], 'direct keys')
        slot.set_value(enable)
        self.midiilluminator.set_enable_direct_keys_for_color(color + 1, enable)
        return True
    
    def __set_enable_direct_notes(self, color, enable):
        slot = self.idx('enable inputs ' + self.__LIGHT_NAME[color], 'direct notes')
        slot.set_value(enable)
        self.midiilluminator.set_enable_direct_notes_for_color(color + 1, enable)
        return True
    
    def __set_enable_midi(self, color, enable):
        slot = self.idx('enable inputs ' + self.__LIGHT_NAME[color], 'midi')
        slot.set_value(enable)
        self.midiilluminator.set_enable_midi_for_color(color + 1, enable)
        return True
    
    def __set_key(self, key, val):
        self.idx('keys')[key].set_value(val)
        self.midiilluminator.set_key(key,val)
        return True
    
    def __set_note(self, note, val):
        self.idx('notes')[note].set_value(val)
        self.midiilluminator.set_note(note,val)
        return True
    
    def __set_channel_color(self, channel, color):
        self.idx('midi channels')[channel].set_value(color)
        self.midiilluminator.set_channel_color(channel, color)
        return True
    
    def __set_note_for_key(self, layout, key, note, repaint=True):
        current_layout = self.idx('key to note maps', 'current layout')
        layouts = self.idx('key to note maps')
        layouts[layout][key].set_value(note)
        active_layout = self.idx('current layout').get_value() + 1
        if layout == active_layout:
            current_layout[key].set_value(note)
        if layout == 1:
           layouts[active_layout][key].set_value(note)
        if (layout == active_layout ) or (layout == self.__idx['key to note maps']['current layout']):
            self.midiilluminator.set_note_for_key(key, note, repaint)
        return True
    
    def __set_enable(self, state):
        self.idx('enable').set_value(state)
        self.midiilluminator.set_enable(state)
        return True
    
    def __set_direct_keys_prio(self, prio):
        self.idx('priorities', 'direct keys').set_value(prio)
        self.midiilluminator.set_direct_keys_prio(prio)
        return True
    
    def __set_direct_notes_prio(self, prio):
        self.idx('priorities', 'direct notes').set_value(prio)
        self.midiilluminator.set_direct_notes_prio(prio)
        return True
    
    def __set_midi_prio(self, prio):
        self.idx('priorities', 'midi').set_value(prio)
        self.midiilluminator.set_midi_prio(prio)
        return True
    
    def _num_keys_in_column(self, column):
        pass
    
    def _num_keys_in_longest_column(self):
        max_length = 0
        for col in range(1,self.num_keys_per_row + 1):
            curr_length = self._num_keys_in_column(col)
            if curr_length > max_length:
                max_length = curr_length
        return max_length
    
    def _sum_keys_in_columns_before_column(self, column):
        sum = 0
        for col in range(1,column):
            sum = sum + self._num_keys_in_column(col)
        return sum
    
    def __export_key_table(self, csvExporter, slot, table_name):
        num_keys_in_longest_column = self._num_keys_in_longest_column()
        row_to_write = [table_name + '[' + `self.num_keys_per_row` + '-' + `num_keys_in_longest_column` + ']']
        curr_key_sum = 0
        for col in range(1,self.num_keys_per_row + 1):
            row_to_write.append(curr_key_sum)
            curr_key_sum = curr_key_sum + self._num_keys_in_column(col)
        csvExporter.writerow(row_to_write)
        for row in range(1, num_keys_in_longest_column + 1):
            row_to_write = [row]
            for col in range(1,self.num_keys_per_row + 1):
                if row <= self._num_keys_in_column(col):
                    if self._sum_keys_in_columns_before_column(col) + row in slot:
                        row_to_write.append(slot[self._sum_keys_in_columns_before_column(col) + row].get_value())
                    else:
                        row_to_write.append(0)
                else:
                    row_to_write.append('#')
            csvExporter.writerow(row_to_write)
    
    def __export(self, do_it):
        if not do_it:
            return True
        self.idx('export csv').set_value(False)
        #FSTDEBUG = self.idx('export csv').get_value()
        #print "FSTDEBUG xas = " + `FSTDEBUG`
        agent_name = self.get_property_string('name')
        config_file = resource.user_resource_file('midi_illuminator',agent_name + '.csv',version='')
        csvExporter = csv.writer(open(config_file, 'wb'), dialect='excel')
        csvExporter.writerow(['sep=',''])
        csvExporter.writerow(['#' + self.get_property_string('name') + ' configuration'])
        csvExporter.writerow(['',])
        csvExporter.writerow(['enable', int(self.idx('enable').get_value())])
        csvExporter.writerow(['transpose', self.idx('transpose').get_value()])
        csvExporter.writerow(['current layout', self.idx('current layout').get_value()])
        csvExporter.writerow(['',])
        csvExporter.writerow(['priorities/'])
        csvExporter.writerow(['direct keys', self.idx('priorities', 'direct keys').get_value()])
        csvExporter.writerow(['direct notes', self.idx('priorities', 'direct notes').get_value()])
        csvExporter.writerow(['midi', self.idx('priorities', 'midi').get_value()])
        csvExporter.writerow(['',])
        csvExporter.writerow(['enable inputs green/'])
        csvExporter.writerow(['direct keys', int(self.idx('enable inputs green', 'direct keys').get_value())])
        csvExporter.writerow(['direct notes', int(self.idx('enable inputs green', 'direct notes').get_value())])
        csvExporter.writerow(['midi', int(self.idx('enable inputs green', 'midi').get_value())])
        csvExporter.writerow(['',])
        csvExporter.writerow(['enable inputs orange/'])
        csvExporter.writerow(['direct keys', int(self.idx('enable inputs orange', 'direct keys').get_value())])
        csvExporter.writerow(['direct notes', int(self.idx('enable inputs orange', 'direct notes').get_value())])
        csvExporter.writerow(['midi', int(self.idx('enable inputs orange', 'midi').get_value())])
        csvExporter.writerow(['',])
        csvExporter.writerow(['enable inputs red/'])
        csvExporter.writerow(['direct keys', int(self.idx('enable inputs red', 'direct keys').get_value())])
        csvExporter.writerow(['direct notes', int(self.idx('enable inputs red', 'direct notes').get_value())])
        csvExporter.writerow(['midi', int(self.idx('enable inputs red', 'midi').get_value())])
        csvExporter.writerow(['',])
        csvExporter.writerow(['midi channels[4-4]',0,4,8,12])
        for i in range(1,5):
            csvExporter.writerow([i,self.idx('midi channels')[i].get_value(),self.idx('midi channels')[i+4].get_value(),
                                  self.idx('midi channels')[i+8].get_value(),self.idx('midi channels')[i+12].get_value()])
        csvExporter.writerow(['',])
        self.__export_key_table(csvExporter, self.idx('keys'), 'keys')
        csvExporter.writerow(['',])
        csvExporter.writerow(['notes[11-12]',0,1,2,3,4,5,6,7,8,9,10])
        note_names = ['c', 'cis', 'd', 'dis', 'e', 'f', 'fis', 'g', 'gis', 'a', 'ais', 'b']
        for note in range(1,13):
            direct_notes_table_row = [note_names[note - 1]]
            for i in range(1,12):
                if note + (i - 1) * 12 < 129:
                    direct_notes_table_row.append(self.idx('notes')[note * i].get_value())
                else:
                    direct_notes_table_row.append('#')
            csvExporter.writerow(direct_notes_table_row)
        csvExporter.writerow(['',])
        csvExporter.writerow(['key to note maps/'])
        for layout in range(1,self._num_layouts() + 1): 
            self.__export_key_table(csvExporter, self.idx('key to note maps', 'layout ' + `layout`), 'layout ' + `layout`)             
        return True
    
    
    def __import(self, do_it):
        if not do_it:
            return True
        if self.is_lite():
            self._init_lite_slots()
        self.idx('import csv').set_value(False)
        config_file = resource.user_resource_file('midi_illuminator',self.get_property_string('name') + '.csv',version='')
        csv_file = open(config_file, 'rb')
        csv_dialect = csv.Sniffer().sniff(csv_file.read(1024))
        csv_file.seek(0)
        csvImporter = csv.reader(csv_file, csv_dialect)
        curr_row = 0
        curr_col = 0
        class states:
            reading_param_name_or_single_value = 0
            reading_table = 1
        curr_slot = self
        last_slots = [curr_slot]
        curr_index = self.__idx
        last_indices = [curr_index]
        table_width = None
        table_height = None
        last_row_with_table_header = 0
        state = states.reading_param_name_or_single_value
        curr_par = None
        enter_group = False
        for row in csvImporter:
            curr_row = curr_row + 1
            curr_col = 0
            for cell in row:
                curr_col = curr_col + 1
                if state == states.reading_table and curr_row - last_row_with_table_header > table_height:
                    state = states.reading_param_name_or_single_value
                    assert len(last_indices) > 0
                    curr_index = last_indices.pop()
                    assert len(last_slots) > 0
                    curr_slot = last_slots.pop()
                    table_width = None
                    table_height = None
                    print "FSTDEBUG leave group"
                if cell == '' or cell.startswith('#') or cell.startswith('sep='):
                    if curr_col == 1:
                        curr_slot = self
                        curr_index = self.__idx
                        curr_par = None
                        table_width = None
                        table_height = None
                        enter_group = False
                        print "FSTDEBUG reset"
                        state = states.reading_param_name_or_single_value
                elif state == states.reading_param_name_or_single_value:
                    if curr_col == 1:
                        cell_match = re.match('([\w\d\s]+)(\/?)',cell)
                        if cell_match <> None:
                            curr_par = cell_match.group(1)
                            if cell_match.group(2) == '/':
                                enter_group = True
                                print "FSTDEBUG: enter_group"
                            print "FSTDEBUG par = " + `curr_par`
                            print "FSTDEBUG curr_index = " + `curr_index`
                            matching_params = filter(lambda par_name: par_name == curr_par, curr_index.keys())
                            if len(matching_params) > 0:
                                cell_match = re.match('([\w\d\s]+)\[(\d+)\-(\d+)\]',cell)
                                if cell_match <> None or enter_group == True:
                                    enter_group = False
                                    last_indices.append(curr_index)
                                    last_slots.append(curr_slot)
                                    if isinstance(curr_index, dict):
                                        curr_index = curr_index[curr_par]
                                        id = None
                                        if isinstance(curr_index, dict):
                                            id = curr_index['id']
                                        else:
                                            id = curr_index
                                            curr_slot = curr_slot[id]
                                    else:
                                        curr_index = {}
                                    print "FSTDEBUG table: new index: " + `curr_index`
                                    if cell_match <> None:
                                        table_width = int(cell_match.group(2))
                                        table_height = int(cell_match.group(3))
                                        state = states.reading_table
                                        last_row_with_table_header = curr_row
                            else:
                                print 'unknown parameter: ' + `cell`
                        else:
                            print 'could not parse ' + `cell`
                    else: # set single value parameter
                        print "FSTDEBUG curr_index = " + `curr_index`
                        if curr_par <> None:
                            id = None
                            if isinstance(curr_index[curr_par], dict):
                                id = curr_index[curr_par]['id']
                            else:
                                id = curr_index[curr_par]
                            curr_slot[curr_index[curr_par]['id']].set_value(int(cell))
                elif state == states.reading_table:
                    if curr_col <> 1 and curr_row > last_row_with_table_header and cell <> '#':
                        row_in_table = curr_row - last_row_with_table_header
                        col_in_table = curr_col - 2
                        print "FSTDEBUG row_in_table = " + `row_in_table` + " col_in_table = " + `col_in_table` + " table_height = " + `table_height` + " cell = " + `cell`
                        #TODO: find more generic criterium
                        if curr_par.startswith('layout') or curr_par.startswith('keys'):
                            curr_slot[self._sum_keys_in_columns_before_column(col_in_table + 1) + row_in_table].set_value(int(cell))
                        else:
                            curr_slot[table_height * col_in_table + row_in_table].set_value(int(cell))
        self.update_native()
        return True
    
    def _init_lite_slots(self):
        self.set_idx('keys',{})
        for key in range(1,self.__num_keys + 1):
            self.idx('keys')[key] = 0
        notenum = 0
        self._lite_slots['notes'] = {}
        for octave in range(0,11):
            for note in ['c', 'cis', 'd', 'dis', 'e', 'f', 'fis', 'g', 'gis', 'a', 'ais', 'b']:
                if notenum > 127:
                    break
                notenum = notenum + 1
                self.idx('notes')[notenum] = 0
        self.set_idx('midi channels', {})
        for channel in range(1,17):
            self.idx('midi channels')[channel] = 0
        self.set_idx_sub('key to note maps', 'current layout', {})
        for key in range(1,self.__num_keys + 1):
            self.idx('key to note maps', 'current layout')[key] = self._default_note(key)
        for layout in range(2,self._num_layouts() + 2):
             self._lite_slots['key to note maps'][layout] = {}
             for key in range(1,self.__num_keys + 1):
                     self.idx('key to note maps')[layout][key] = self._default_note(key)
        self.set_idx_sub('priorities','direct keys', 3)
        self.set_idx_sub('priorities','direct notes', 2)
        self.set_idx_sub('priorities','midi', 1)

    def update_native(self):
        self.__set_transpose(self.idx('transpose').get_value())
        self.__set_layout(self.idx('current layout').get_value(), force=True)
        for color in range(0,3):
            self.__set_enable_direct_keys(color, bool(self.idx('enable inputs ' + self.__LIGHT_NAME[color],'direct keys').get_value()))
            self.__set_enable_direct_notes(color, bool(self.idx('enable inputs ' + self.__LIGHT_NAME[color],'direct notes').get_value()))
            self.__set_enable_midi(color, bool(self.idx('enable inputs ' + self.__LIGHT_NAME[color],'midi').get_value()))
    
        for key in range(1,self.__num_keys + 1):
            self.__set_key(key, self.idx('keys')[key].get_value())
            
        for note in range(1,129):
            self.__set_note(note, self.idx('notes')[note].get_value())
    
        for channel in range(1,17):
            self.__set_channel_color(channel, self.idx('midi channels')[channel].get_value())
    
        self.__set_enable(bool(self.idx('enable').get_value()))
        
        self.__set_direct_keys_prio(self.idx('priorities','direct keys').get_value())
        self.__set_direct_notes_prio(self.idx('priorities','direct notes').get_value())
        self.__set_midi_prio(self.idx('priorities','midi').get_value())
    
        return True
