
/*
 Copyright 2009 Eigenlabs Ltd.  http://www.eigenlabs.com

 This file is part of EigenD.

 EigenD is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 EigenD is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with EigenD.  If not, see <http://www.gnu.org/licenses/>.
*/

//#pragma optimize( "", off )

#include "midiilluminator.h"
#include "simpleagent.h"

#include <piw/piw_clock.h>
#include <piw/piw_tsd.h>
#include <picross/pic_ilist.h>
#include <cmath>
#include <map>

#define ENTER_TRACE(method_name) 	pic::logmsg() << "enter " << method_name;
//#define ENTER_TRACE(method_name)

using namespace midiilluminator_plg;

typedef unsigned color_t;
typedef unsigned note_t;
typedef unsigned keys_t;
typedef unsigned prio_t;
typedef unsigned channel_t;
typedef signed semitones_offset_t;

struct note_and_channel_t
{
	note_t note;
	channel_t channel;
	bool operator==(const note_and_channel_t& comp)
	{
		return (comp.note == note) && (comp.channel == channel);
	}
};

namespace
{
	struct midiilluminator_wire_t
		: simpleagent_wire_t
	{
		midiilluminator_wire_t(midiilluminator_t::impl_t *p, const piw::event_data_source_t& es, unsigned num_keys);
	public:
		virtual ~midiilluminator_wire_t();
		void process(unsigned, const piw::data_nb_t &, unsigned long long);
		void refresh_led_matrix();
		void send_led_matrix();
		unsigned char* led_matrix_;
		unsigned num_keys_;
	};
}
struct midiilluminator_t::impl_t
	: simpleagent_impl_t
{
	impl_t(piw::clockdomain_ctl_t *cd, const piw::cookie_t &c, const unsigned& num_keys);
	unsigned num_keys_;
	virtual simpleagent_wire_t* create_wire(const piw::event_data_source_t& es);
	DECLARE_FASTCALL_PROPERTY(semitones_offset_t, transpose)
	DECLARE_FASTCALL_PROPERTY(bool, enable)
	DECLARE_FASTCALL_PROPERTY(prio_t, midi_prio)
	DECLARE_FASTCALL_PROPERTY(prio_t, direct_notes_prio)
	DECLARE_FASTCALL_PROPERTY(prio_t, direct_keys_prio)
	DECLARE_FASTCALL_PROPERTY_MAP(color_t, enable_direct_keys_for_color, bool, enable)
	DECLARE_FASTCALL_PROPERTY_MAP(color_t, enable_direct_notes_for_color, bool, enable)
	DECLARE_FASTCALL_PROPERTY_MAP(color_t, enable_midi_for_color, bool, enable)
	DECLARE_FASTCALL_PROPERTY_MAP(keys_t, key, color_t, color)
	DECLARE_FASTCALL_PROPERTY_MAP(note_t, note, color_t, color)
	DECLARE_FASTCALL_PROPERTY_MAP(channel_t, channel_color, color_t, color)
	//DECLARE_FASTCALL_PROPERTY_MAP(keys_t, note_for_key, note_t, note)
	std::map<keys_t,note_t> note_for_key_map_;
	static int set_note_for_key_static(void* i, void* par);
	void set_note_for_key(const keys_t key, const note_t note, bool refresh);
	//DECLARE_FASTCALL_PROPERTY_MAP END

	bool refresh_color_for_keys_from_midi(note_t note, channel_t channel, bool isNoteOn);
	bool refresh_color_for_keys_from_direct_note(note_t note, color_t color);
	void trigger_refresh_matrix();
	note_t transpose(note_t note);
	std::map<note_t,std::list<keys_t> > keys_for_note_;
	std::map<keys_t, color_t> color_for_key_from_midi_;
	std::map<keys_t, color_t> color_for_key_from_direct_notes_;
	std::list<note_and_channel_t> active_midi_notes_;
};

// from plg_midi/midi_input.cpp - we interpret c as channel for the key events - see below
const unsigned NOTE_ON = 144;
const unsigned NOTE_OFF = 128;
const unsigned MIDI_COMMAND_MASK = 0xF0;
const unsigned MIDI_CHANNEL_MASK = 0xF;

struct blob_t
{
    unsigned char command_and_channel,note,velocity;
};

const color_t color_disabled = 0;
const color_t color_green = 1;
const color_t color_orange = 2;
const color_t color_red = 3;

//midiilluminator_t implementation

midiilluminator_t::midiilluminator_t(piw::clockdomain_ctl_t *cd, const piw::cookie_t &cookie, const unsigned& num_keys)
	: impl_(new impl_t(cd,cookie,num_keys))
{
	ENTER_TRACE("midiilluminator_t constructor")
}

midiilluminator_t::~midiilluminator_t()
{
	ENTER_TRACE("midiilluminator_t destructor")
	delete impl_;
}

piw::cookie_t midiilluminator_t::cookie()
{
	ENTER_TRACE("midiilluminator_t cookie")
	return piw::cookie_t(impl_);
}


// midiilluminator_wire_t implementation

midiilluminator_wire_t::midiilluminator_wire_t(midiilluminator_t::impl_t *p, const piw::event_data_source_t& es, unsigned num_keys)
	: simpleagent_wire_t(p,es,1), num_keys_(num_keys)
{
	ENTER_TRACE("midiilluminator_wire_t constructor")
	led_matrix_ = new unsigned char[num_keys_];
	memset(led_matrix_, 0, num_keys_);
}

midiilluminator_wire_t::~midiilluminator_wire_t()
{
	ENTER_TRACE("midiilluminator_wire_t destructor")
	delete[] led_matrix_;
}

void midiilluminator_wire_t::process(unsigned input_signal, const piw::data_nb_t& data, unsigned long long timestamp)
{
	ENTER_TRACE("midiilluminator_wire_t process")
	pic::logmsg() << "input_signal" << input_signal << " timestamp: " << timestamp;
	for(unsigned int i = 0; i < data.as_bloblen();++i)
		pic::logmsg() << "data[" << i <<"]: " << (int)((const unsigned char*)data.as_blob())[i];
	if(input_signal == 1) // we expect note data only on signal 1
	{
		blob_t* midi_msg = (blob_t*)data.as_blob();
		bool isNoteOn = (midi_msg->command_and_channel & MIDI_COMMAND_MASK) == NOTE_ON;
		bool isNoteOff = (midi_msg->command_and_channel & MIDI_COMMAND_MASK) == NOTE_OFF;
 		if(isNoteOn || isNoteOff)
		{
			// *hack* we expect the channel to be sent in
			// blob_t::c for key events. As the original midi input agent always
			// sends 0 here we will always get channel 1 in that case.
			// a "hacked" midi input can send us different channels here
			// until a proper solution is found
			midiilluminator_t::impl_t* root = (midiilluminator_t::impl_t*)root_;
 			channel_t channel = (midi_msg->command_and_channel & MIDI_CHANNEL_MASK) + 1;
			note_t note = midi_msg->note + 1; //1-128: notes; 0 no note (necessary for key to note map)
			note_and_channel_t note_and_channel = {note, channel};
			if(isNoteOn)
				root->active_midi_notes_.push_back(note_and_channel);
			else
				root->active_midi_notes_.remove(note_and_channel);
			bool color_changed = root->refresh_color_for_keys_from_midi(root->transpose(note), channel, isNoteOn);
			if(color_changed)
				refresh_led_matrix();
		}
	}
}

void midiilluminator_wire_t::refresh_led_matrix()
{
	ENTER_TRACE("midiilluminator_wire_t refresh_led_matrix")
	bool colors_changed = false;
	midiilluminator_t::impl_t* root = static_cast<midiilluminator_t::impl_t*>(root_);
	for(unsigned curr_key = 1; curr_key < num_keys_ + 1; ++curr_key)
	{
		// smaller prio number means higher priority
		if(root->enable_)
		{
			color_t new_color = color_disabled;
			for(prio_t prio = 1; prio < 4; ++prio)
			{
				color_t key_color = color_disabled;
				if(root->midi_prio_ <= prio)
				{
					color_t midi_color = root->color_for_key_from_midi_[curr_key];
					if(root->enable_midi_for_color_map_[midi_color])
						key_color = midi_color;
				}
				if((root->direct_notes_prio_ <= prio) && (key_color == color_disabled))
				{
					color_t direct_notes_color = root->color_for_key_from_direct_notes_[curr_key];
					if(root->enable_direct_notes_for_color_map_[direct_notes_color])
						key_color = direct_notes_color;
				}
				if((root->direct_keys_prio_ <= prio) && (key_color == color_disabled))
				{
					color_t direct_keys_color = root->key_map_[curr_key];
					if(root->enable_direct_keys_for_color_map_[direct_keys_color])
						key_color = direct_keys_color;
				}
				if(key_color != color_disabled)
				{
					new_color = key_color;
					break;
				}
			}
			if(led_matrix_[curr_key - 1] != new_color)
			{
				led_matrix_[curr_key - 1] = new_color;
				colors_changed = true;
			}
		}
		else // !enable_
		{
			if(led_matrix_[curr_key - 1] != 0)
			{
				led_matrix_[curr_key - 1] = 0;
				colors_changed = true;
			}
		}
	}
	if(colors_changed)
		send_led_matrix();
}

void midiilluminator_wire_t::send_led_matrix()
{
	ENTER_TRACE("midiilluminator_wire_t send_led_matrix")
	unsigned char* raw_data = NULL;
	piw::data_nb_t led_data = piw::makeblob_nb(0, num_keys_, &raw_data);
	memcpy(raw_data, led_matrix_, num_keys_);
	output_.add_value(1,led_data);
}


// midiilluminator_t::impl_t implementation
midiilluminator_t::impl_t::impl_t(piw::clockdomain_ctl_t *cd, const piw::cookie_t &c, const unsigned& num_keys)
	: simpleagent_impl_t(cd,c, "midi illuminator"), num_keys_(num_keys),
		transpose_(0),  enable_(true), midi_prio_(1), direct_notes_prio_(2), direct_keys_prio_(3)
{
	ENTER_TRACE("midiilluminator_t::impl_t constructor")
	enable_direct_keys_for_color_map_[color_green] = true;
	enable_direct_keys_for_color_map_[color_orange] = true;
	enable_direct_keys_for_color_map_[color_red] = true;
	enable_direct_notes_for_color_map_[color_green] = true;
	enable_direct_notes_for_color_map_[color_orange] = true;
	enable_direct_notes_for_color_map_[color_red] = true;
	enable_midi_for_color_map_[color_green] = true;
	enable_midi_for_color_map_[color_orange] = true;
	enable_midi_for_color_map_[color_red] = true;
}

DEFINE_FASTCALL_PROPERTY(midiilluminator, int, transpose)
DEFINE_FASTCALL_PROPERTY(midiilluminator, bool, enable)
DEFINE_FASTCALL_PROPERTY(midiilluminator, prio_t, direct_keys_prio)
DEFINE_FASTCALL_PROPERTY(midiilluminator, prio_t, direct_notes_prio)
DEFINE_FASTCALL_PROPERTY(midiilluminator, prio_t, midi_prio)
DEFINE_FASTCALL_PROPERTY_MAP(midiilluminator, color_t, enable_direct_keys_for_color, bool, enable)
DEFINE_FASTCALL_PROPERTY_MAP(midiilluminator, color_t, enable_direct_notes_for_color, bool, enable)
DEFINE_FASTCALL_PROPERTY_MAP(midiilluminator, color_t, enable_midi_for_color, bool, enable)
DEFINE_FASTCALL_PROPERTY_MAP(midiilluminator, keys_t, key, color_t, color)
DEFINE_FASTCALL_PROPERTY_MAP(midiilluminator, unsigned, note, color_t, color)
DEFINE_FASTCALL_PROPERTY_MAP(midiilluminator, note_t, channel_color, color_t, color)
//DEFINE_FASTCALL_PROPERTY_MAP(midiilluminator, keys_t, note_for_key, unsigned, note)
struct set_note_for_key_params
{
	set_note_for_key_params(const keys_t& k, unsigned n, bool r) : key(k), note(n), refresh(r) {}
	keys_t key;
	unsigned note;
	bool refresh;
};
int midiilluminator_t::impl_t::set_note_for_key_static(void* i, void* par)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_note_for_key_static")
	midiilluminator_t::impl_t *i_ = (midiilluminator_t::impl_t *)i;
	set_note_for_key_params par_ = *(set_note_for_key_params*)par;
	i_->set_note_for_key(par_.key, par_.note, par_.refresh);
	return 0;
}
void midiilluminator_t::set_note_for_key(keys_t key, unsigned note, bool refresh)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_note_for_key")
	set_note_for_key_params pars(key, note, refresh);		\
	piw::tsd_fastcall(midiilluminator_t::impl_t::set_note_for_key_static, impl_, &pars);
}
//DEFINE_FASTCALL_PROPERTY_MAP END

simpleagent_wire_t* midiilluminator_t::impl_t::create_wire(const piw::event_data_source_t& es) \
{
	ENTER_TRACE("midiilluminator_t::impl_t create_wire")
	return new midiilluminator_wire_t(this, es, num_keys_);
}

note_t midiilluminator_t::impl_t::transpose(note_t note)
{
	ENTER_TRACE("midiilluminator_t::impl_t transpose")
	note_t transposed_note = note + transpose_;
	return transposed_note < 128?transposed_note:0;
}


bool midiilluminator_t::impl_t::refresh_color_for_keys_from_midi(note_t note, channel_t channel, bool isNoteOn)
{
	ENTER_TRACE("midiilluminator_t::impl_t refresh_color_for_keys_from_midi")
	bool color_changed = false;
	color_t color_for_curr_note = isNoteOn?channel_color_map_[channel]:0;
	std::list<keys_t>& keys_for_current_note = keys_for_note_[note];
	for(std::list<keys_t>::iterator key_iter = keys_for_current_note.begin();
			key_iter != keys_for_current_note.end(); ++key_iter)
	{
		if(color_for_key_from_midi_[*key_iter] != color_for_curr_note)
		{
			color_changed = true;
			color_for_key_from_midi_[*key_iter] = color_for_curr_note;
		}
	}
	return color_changed;
}

bool midiilluminator_t::impl_t::refresh_color_for_keys_from_direct_note(note_t note, color_t color_for_curr_note)
{
	ENTER_TRACE("midiilluminator_t::impl_t refresh_color_for_keys_from_direct_note")
	bool color_changed = false;
	std::list<keys_t>& keys_for_current_note = keys_for_note_[note];
	for(std::list<keys_t>::iterator key_iter = keys_for_current_note.begin();
			key_iter != keys_for_current_note.end(); ++key_iter)
	{
		if(color_for_key_from_direct_notes_[*key_iter] != color_for_curr_note)
		{
			color_changed = true;
			color_for_key_from_direct_notes_[*key_iter] = color_for_curr_note;
		}
	}
	return color_changed;
}

void midiilluminator_t::impl_t::trigger_refresh_matrix()
{
	ENTER_TRACE("midiilluminator_t::impl_t trigger_refresh_matrix")
	pic::weak_t<simpleagent_wire_t> current_wire_weak(current_wire_);
	midiilluminator_wire_t* current_wire = static_cast<midiilluminator_wire_t*>(current_wire_weak.ptr());
	if(current_wire)
	{
		current_wire->refresh_led_matrix();
	}
}

void midiilluminator_t::impl_t::set_note_for_key(const unsigned key, const unsigned note, bool refresh)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_note_for_key")
	if((note_for_key_map_.count(key) == 0) || (note_for_key_map_[key] != note))
	{
		if(note_for_key_map_.count(key) > 0)
			keys_for_note_[note_for_key_map_[key]].remove(key); // key is not mapped to old note anymore so remove old note from list of notes mapped to this key
		keys_for_note_[note].push_back(key); // map key to new note
		note_for_key_map_[key] = note;
	}
	if(refresh)
		set_transpose(transpose_);
}

void midiilluminator_t::impl_t::set_channel_color(const unsigned& channel, const unsigned& color)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_channel_color")
	if(channel_color_map_[channel] != color)
	{
		channel_color_map_[channel] = color;
		bool color_changed = false;
		for(std::list<note_and_channel_t>::iterator active_midi_notes_iter = active_midi_notes_.begin();
				active_midi_notes_iter != active_midi_notes_.end();++active_midi_notes_iter)
		{
			note_t transposed_note = transpose(active_midi_notes_iter->note);
			bool refresh_changed_color = refresh_color_for_keys_from_midi(transposed_note, active_midi_notes_iter->channel, true);
			if(refresh_changed_color)
				color_changed = true;
		}
		if(color_changed)
			trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_note(const unsigned& note, const unsigned& color)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_note")
	if(note_map_[note] != color)
	{
		note_map_[note] = color;
		bool color_changed = refresh_color_for_keys_from_direct_note(transpose(note), color);
		if(color_changed)
			trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_key(const unsigned& key, const unsigned& color)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_key")
	if(key_map_[key] != color)
	{
		key_map_[key] = color;
		trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_enable_direct_keys_for_color(const unsigned& color, const bool& enable)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_enable_direct_keys_for_color")
	if(enable_direct_keys_for_color_map_[color] != enable)
	{
		enable_direct_keys_for_color_map_[color] = enable;
		trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_enable_direct_notes_for_color(const unsigned& color, const bool& enable)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_enable_direct_notes_for_color")
	if(enable_direct_notes_for_color_map_[color] != enable)
	{
		enable_direct_notes_for_color_map_[color] = enable;
		trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_enable_midi_for_color(const unsigned& color, const bool& enable)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_enable_midi_for_color")
	if(enable_midi_for_color_map_[color] != enable)
	{
		enable_midi_for_color_map_[color] = enable;
		trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_transpose(const int& transp)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_transpose")
	transpose_ = transp;
	bool color_changed = false;
	color_for_key_from_midi_.clear();
	for(std::list<note_and_channel_t>::iterator active_midi_notes_iter = active_midi_notes_.begin();
			active_midi_notes_iter != active_midi_notes_.end();++active_midi_notes_iter)
	{
		note_t transposed_note = transpose(active_midi_notes_iter->note);
		bool refresh_changed_color = refresh_color_for_keys_from_midi(transposed_note, active_midi_notes_iter->channel, true);
		if(refresh_changed_color)
			color_changed = true;
	}
	color_for_key_from_direct_notes_.clear();
	for(std::map<note_t,color_t>::iterator direct_notes_iter = note_map_.begin();direct_notes_iter != note_map_.end();++direct_notes_iter)
	{
		note_t transposed_note = transpose(direct_notes_iter->first);
		bool refresh_changed_color = refresh_color_for_keys_from_direct_note(transposed_note, direct_notes_iter->second);
		if(refresh_changed_color)
			color_changed = true;
	}
	trigger_refresh_matrix();
}

void midiilluminator_t::impl_t::set_enable(const bool& enable)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_enable")
	if(enable_ != enable)
	{
		enable_ = enable;
		trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_direct_keys_prio(const unsigned& prio)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_direct_keys_prio")
	if(direct_keys_prio_ != prio)
	{
		direct_keys_prio_ = prio;
		trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_direct_notes_prio(const unsigned& prio)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_direct_notes_prio")
	if(direct_notes_prio_ != prio)
	{
		direct_notes_prio_ = prio;
		trigger_refresh_matrix();
	}
}

void midiilluminator_t::impl_t::set_midi_prio(const unsigned& prio)
{
	ENTER_TRACE("midiilluminator_t::impl_t set_midi_prio")
	if(midi_prio_ != prio)
	{
		midi_prio_ = prio;
		trigger_refresh_matrix();
	}
}

//#pragma optimize( "", on )
